/* cart.c */
/* Lego cart main program for download to RCX */
/* Copyright (C) 2003 by Peter J. Gawthrop */

/* Under RCS version control */
/* 	$Id: cart.c,v 1.63 2003/07/25 12:01:12 peterg Exp $	 */

/* Openloop - comment out to use closed-loop */
/* #define OPEN_LOOP 1 */

/* Use cart position -- comment out to use load position */
/* The default is to comment out the following line */
/* #define USE_CART_POSITION  */

#include "cart.h"		/* Main Header file */



time_t sys_time,first_time;	/* System time */
double elapsed;			/* Elapsed time since start */
double U[N_U],     U_new[N_U];
double ustar[N_U];

int i_U=0;

int new_U=0;

#include "sensors.c"
#include "actuators.c"

void set_ustar(double ustar[N_U], double Ustar[N_T][N_U], int i_U)
{
  int i;
  for (i=0;i<N_U;i++)
    ustar[i] = Ustar[i_U][i];
}

void initialise_u(double U[N_U])
{
  int i;
  for (i=0;i<N_U;i++)
    U[i] = 0.0;
}

void initialise_state(double state[N_state])
{
  int i;
  for (i=0;i<N_state;i++)
    state[i] = 0.0;
}

/* Send data as integers via tower */
#include "putint.c"
void put_data(double elapsed, double y, double u, double v, double w)
{
  int I[N_data];

  I[0] = elapsed*100;
  I[1] = y*I2R;
  I[2] = u*I2R;
  I[3] = v*I2R;
  I[4] = w*I2R;

  putint(I);
  
};

void get_U(const unsigned char *data, unsigned char len)
/* Gets the U vector from laptop via IR */
{
  int i,I;
  char c[2];

/* Get U */
  for (i=0;i<N_U;i++)
    {
      c[0] = data[2*i];
      c[1] = data[2*i+1];
      memcpy(&I,&c,2);
      U_new[i] = I/I2R;
    };

  new_U = 1;
};

#include "controller.c"		/* User-supplied controller code */

int main(int argc, char **argv) 
{

  int i;

  double kk;
  double bob, angle, angle_0, position, position_old, velocity;
  double w = 0.0; 		/* metre */
  double y, y_con, u, u_out;
  double state[N_state];	/* Controller state */
  double t=0.0;

  sleep(1);
  lcd_clear();

  /* Initialise */
  new_U=0;
  i_U = 0;
  initialise_sensors(); 
  initialise_u(U);
  initialise_u(U_new);
  initialise_state(state);
  set_ustar(ustar,Ustar,i_U);

  cputs("WAIT");

  /* Setup lnp from laptop */
  lnp_integrity_set_handler(get_U);
  cputs("START");
  first_time = sys_time;

  /* Main loop */
  kk = 0;
  angle_0 = pendulum_angle(0.0);
  position_old = cart_position();
  u_out = 0.0;
  u = 0.0;

  while(1)			/* Loop forever */
    { 
      angle = pendulum_angle(angle_0);
      bob = bob_position(angle);
      position = cart_position();
      velocity = motor_velocity();
      y = position+bob;		/* Output */

#ifdef USE_CART_POSITION 
      y_con = position;		/* Controlled output */
#else
      y_con = y;		/* Controlled output */
#endif

      elapsed = (sys_time-first_time)/1000.0;

      /* Handle the new value of U */
      if (new_U==1)
	{
	  for (i=0;i<N_U;i++)
	    U[i] = U_new[i];	/* Change U here */
	  new_U = 0;
	  i_U = 0;
	  /* Reinitialise ustar */
	  set_ustar(ustar,Ustar,i_U);
	};

      /* Compute OL setpoint for local controller */
      w = 0;
      for (i=0;i<N_U;i++)
	{
	  w = w + U[i]*ustar[i];
	};

      lcd_int(i_U);		/* Display status */

      /* Control */
#ifdef OPEN_LOOP
      u = w;
#else
      u = controller(y_con,u_out,w,state);
#endif

      u_out = drive(u);		/* Clip u if required */
      put_data(elapsed,y,u_out,velocity,w); /* Send data to laptop */

      /* Increment the counter for predefined Ustar */
      if (i_U<N_T-1)		/* Don't run off the end */
	i_U++;

      t = t+DT;			/* Increment time */
      position_old = position;	/*Save y for v computation */

      /* Next Ustar vector */
 	  set_ustar(ustar,Ustar,i_U);
    }
     
  return 0;

}
